#!/usr/local/ensembl/bin/perl -w
# written by Steve Searle 
# script calculates the repeat coverage of a genome which is repeatmasked by a 
# specific analysis (used for ciona JGI2 build). 
# output is in format : 
# 
#  Number bases masked = 267592
#  Number genes overlapped = 0
#  Total bases = 173499994
#  Total masked = 23746780
#  Total genes overlapped = 0
#
# to run on an assembly with coord_system.version COD_PRE and only toplevel regions : 
#
# perl repeat_coverage.pl $dbcon jhv_cod1_ref -repeattypes RepeatMask_fish -path COD_PRE -coord_system toplevel 
# 




use strict;
use Bio::EnsEMBL::DBSQL::DBAdaptor;
use Bio::EnsEMBL::SeqFeature;
use Getopt::Long;


my $host   = 'ia64g';
my $user   = 'ensro';
my $dbname ; 
my $port   = 3306;
my $dbpass ; 
my $dnahost   = $host;  
my $dnauser   = $user;
my $dnadbname =$dbname ; 
my $dnaport = 3306;

my @chromosomes;

#my @defgenetypes = ('ensembl');
my @defgenetypes = ();
my @genetypes;
my $schema = 20;
my $coord_system='chromosome' ;

# my @defrepeattypes = ('RepeatMask_piler');

my @defrepeattypes = ();
my @repeattypes;

my $path ; 
my $verbose ; 
$| = 1;

&GetOptions(
  'dbhost|host:s'   => \$host,
  'dbuser|user:s'   => \$user,
  'dbpass:s' => \$dbpass,
  'dbname:s' => \$dbname,
  'path:s'   => \$path,
  'dbport|port:n'   => \$port,
  'dnahost:s'   => \$dnahost,
  'dnauser:s'   => \$dnauser,
  'dnadbname:s' => \$dnadbname,
  'dnaport:n'   => \$dnaport,
  'chromosomes:s' => \@chromosomes,
  'genetypes:s' => \@genetypes,
  'repeattypes:s' => \@repeattypes,
  'coord_system:s' => \$coord_system,
  'schema:n' => \$schema,
  'verbose!'   => \$verbose,
);

if (scalar(@chromosomes)) {
  @chromosomes = split(/,/,join(',',@chromosomes));
}

if (scalar(@genetypes)) {
  @genetypes = split(/,/,join(',',@genetypes));
} else {
  @genetypes = @defgenetypes;
}

if (scalar(@repeattypes)) {
  @repeattypes = split(/,/,join(',',@repeattypes));
} else {
  @repeattypes = @defrepeattypes;
}

print join (" ", @repeattypes) ; 

my $db = new Bio::EnsEMBL::DBSQL::DBAdaptor(
  -host   => $host,
  -user   => $user,
  -pass   => $dbpass,
  -port   => $port,
  -dbname => $dbname
);
#$db->assembly_type($path);

my $dnadb;
if ($dnadbname) {
  $dnadb = new Bio::EnsEMBL::DBSQL::DBAdaptor(
    -host   => $dnahost,
    -user   => $dnauser,
    -port   => $dnaport,
    -dbname => $dnadbname
  );
  #$dnadb->assembly_type($path);

  $db->dnadb($dnadb);
}


my $sa   = $db->get_SliceAdaptor();
my $ga   = $db->get_GeneAdaptor();
my $dafa = $db->get_DnaAlignFeatureAdaptor();
my $chrhash;
if ($schema == 20) {
  $chrhash = get_chrlengths_v20($db, $path, $coord_system);
} else {
  $chrhash = get_chrlengths_v19($db, $path);
}


#filter to specified chromosome names only
if (scalar(@chromosomes)) {
  foreach my $chr (@chromosomes) {
    my $found = 0;
    foreach my $chr_from_hash (keys %$chrhash) {
      if ($chr_from_hash =~ /^${chr}$/) {
        $found = 1;
        last;
      }
    }
    if (!$found) {
      print "Didn't find chromosome named $chr in database $dbname\n";
    }
  }
  HASH: foreach my $chr_from_hash (keys %$chrhash) {
    foreach my $chr (@chromosomes) {
      if ($chr_from_hash =~ /^${chr}$/) {next HASH;}
    }
    delete($chrhash->{$chr_from_hash});
  }
}


$| = 1;

my $total_masked = 0;
my $total_overgene = 0;
my $total_bases = 0; 
CHR:
foreach my $chr (reverse sort bychrnum keys %$chrhash) {
  print STDERR "Chr $chr from 1 to " . $chrhash->{$chr}{'length'}. "\n" if $verbose ; 
  my $chrstart = 1;
  my $chrend   = $chrhash->{$chr}{'length'} ; 
  my $coord_system = $chrhash->{$chr}{'coord'} ; 
  my $slicename = "$coord_system:$path:$chr:" . $chrstart . ":" . $chrend . ":1";
  my $slice;
  if ($schema == 20) {
    $slice = $sa->fetch_by_name($slicename);
  } else {
    $slice = $sa->fetch_by_chr_start_end($chr,$chrstart,$chrend);
  }

  print "Fetching genes\n" if $verbose ;

  my @genes;
  my %genes_hash;
  foreach my $genetype (@genetypes) {
    $genes_hash{$genetype} = $slice->get_all_Genes_by_type($genetype);
    print "Got " . scalar(@{$genes_hash{$genetype}}) . " $genetype genes\n" if $verbose ;
    push @genes,@{$genes_hash{$genetype}};
  }
  print "Done fetching genes (fetched " . scalar(@genes) .")\n" if $verbose ; 

  print "Fetching repeats\n" if $verbose ;
  my @repeats;
  foreach my $repeattype (@repeattypes) { 
    my @rp = @{$slice->get_all_RepeatFeatures($repeattype)}; 
    push @repeats,@rp ; 
    print scalar(@rp ) . " repeatas found -  $repeattype --> " . $slice->name . "\n" if $verbose ; 
  }
  print "Done fetching repeats (fetched " . scalar(@repeats) .")\n" if $verbose ; 

  @repeats = map { $_->[1] } sort { $a->[0] <=> $b->[0] } map { [$_->start, $_] } @repeats;

  my @repeat_blocks;

  my $curblock = undef;
  REPLOOP: foreach my $repeat (@repeats) {
    if ($repeat->start <= 0) { $repeat->start(1); }
    if (defined($curblock) && $curblock->end >= $repeat->start) {
      #print "Adding to block with " . $repeat->start . " to " . $repeat->end . "\n";
      #print "          Block was " . $curblock->start . " to " . $curblock->end . "\n";
      if ($repeat->end > $curblock->end) { $curblock->end($repeat->end); }
    } else {
      #print "Starting new block with " . $repeat->start . " to " . $repeat->end . "\n";
      $curblock = Bio::EnsEMBL::SeqFeature->new(-START => $repeat->start, -END => $repeat->end);
      push @repeat_blocks,$curblock;
    }
  }

  @repeat_blocks = map { $_->[1] } sort { $a->[0] <=> $b->[0] } map { [$_->start, $_] } @repeat_blocks;

  @genes = sort {$a->start <=> $b->start} @genes;
  my $ngene = scalar(@genes);


  my $novergene = 0;

  my $nmasked = 0;
  foreach my $block (@repeat_blocks) {
    $nmasked += $block->length;

   GENE: 
    for (my $ind=0; $ind < $ngene; $ind++) {
      my $gene = $genes[$ind];
      if (!defined($gene)) {
        next GENE;
      }
      if ($block->end >= $gene->start &&
        $block->start <= $gene->end) {
        foreach my $exon (@{$gene->get_all_Exons}) {
          if ($exon->overlaps($block)) {
            print "Overlap for gene " . get_gene_id($gene) . "\n" if $verbose ; 
            $novergene++;
            $genes[$ind] = undef;
            next GENE;
          }  
        }
      } elsif ($gene->start > $block->end) {
        last;
      } elsif ($gene->end < $block->start) {
        #print "Removing gene " . $genes[$ind]->stable_id . "\n";
        $genes[$ind] = undef;
      }
    }
  }
  print "Number bases masked = $nmasked\n" if $verbose ; 
  print "Number genes overlapped = $novergene\n" if $verbose ;
 
  $total_overgene += $novergene;
  $total_masked += $nmasked;
  $total_bases += $slice->length;
}
print "Total bases = $total_bases\n";
print "Total masked = $total_masked\t" ; 
my $ratio = ($total_masked / $total_bases) * 100  ; 
print " ( $ratio % masked) \n" ; 
print "Total genes overlapped = $total_overgene\n";
print "Done\n";


sub get_gene_id {
  my ($gene) = @_;

  return $gene->stable_id . " (" . $gene->dbID . ")" if ($gene->stable_id);

  return $gene->dbID;
}

sub get_transcript_id {
  my ($transcript) = @_;

  return $transcript->stable_id if ($transcript->stable_id);

  return $transcript->dbID;
}

sub get_exon_id {
  my ($exon) = @_;

  return $exon->stable_id if ($exon->stable_id);

  return $exon->dbID;
}

sub bychrnum {

  my @awords = split /_/, $a;
  my @bwords = split /_/, $b;

  my $anum = $awords[0];
  my $bnum = $bwords[0];

  #  if ($anum !~ /^chr/ || $bnum !~ /^chr/) {
  #    die "Chr name doesn't begin with chr for $a or $b";
  #  }

  $anum =~ s/chr//;
  $bnum =~ s/chr//;

  if ($anum !~ /^[0-9]*$/) {
    if ($bnum !~ /^[0-9]*$/) {
      return $anum cmp $bnum;
    } else {
      return 1;
    }
  }
  if ($bnum !~ /^[0-9]*$/) {
    return -1;
  }

  if ($anum <=> $bnum) {
    return $anum <=> $bnum;
  } else {
    if ($#awords == 0) {
      return -1;
    } elsif ($#bwords == 0) {
      return 1;
    } else {
      return $awords[1] cmp $bwords[1];
    }
  }
}

sub get_chrlengths_v20 {
  my $db   = shift;
  my $type = shift;
  my $coord_system = shift;

  if (!defined($coord_system)) {
    $coord_system = 'chromosome';
  }

  if (!$db->isa('Bio::EnsEMBL::DBSQL::DBAdaptor')) {
    die "get_chrlengths should be passed a Bio::EnsEMBL::DBSQL::DBAdaptor\n";
  }

  my $query ;
 
  if ( $coord_system =~/toplevel/ ) { 
      $query  = " select s.name , s.length as mce , c.name as mce_name from seq_region s, seq_region_attrib sa  , " . 
                  " coord_system c , attrib_type at where s.seq_region_id = sa.seq_region_id " . 
                  " and sa.attrib_type_id = at.attrib_type_id and at.code ='".$coord_system.
                  "' and s.coord_system_id = c.coord_system_id and c.version='".$type. "'" ; 

  } else { 
    $query = "select seq_region.name, seq_region.length as mce , coord_system.name  as mce_name  from seq_region,coord_system  where" .
              " seq_region.coord_system_id=coord_system.coord_system_id and" .
              " coord_system.version = '" . $type . "' and coord_system.name='" . $coord_system . "'"; 

 }   

  my $sth = $db->prepare($query);

  print "\nQuery : $query\n" ; 
  $sth->execute;

  my %chrhash;

  my $hashref;
  while (($hashref = $sth->fetchrow_hashref) && defined($hashref)) {  
   # foreach ( keys %$hashref ) {  
   #     print "$_ --> $$hashref{$_} \n" ; 
   # }  
    my $contig_name = $hashref->{'name'} ; 
    $chrhash{$contig_name}{'length'} = $hashref->{mce};
    $chrhash{$contig_name}{'coord'} = $hashref->{mce_name};
  }
  return \%chrhash;
}


sub get_chrlengths_v19 {
  my $db = shift;
  my $type = shift;

  if (!$db->isa('Bio::EnsEMBL::DBSQL::DBAdaptor')) {
    die "get_chrlengths should be passed a Bio::EnsEMBL::DBSQL::DBAdaptor\n";
  }

  my %chrhash;

  my $q = qq( SELECT chrom.name,max(chr_end) FROM assembly as ass, chromosome as chrom
              WHERE ass.type = '$type' and ass.chromosome_id = chrom.chromosome_id
              GROUP BY chrom.name
            );

  my $sth = $db->prepare($q) || $db->throw("can't prepare: $q");
  my $res = $sth->execute || $db->throw("can't execute: $q");

  while( my ($chr, $length) = $sth->fetchrow_array) {
    $chrhash{$chr} = $length;
  }
  return \%chrhash;
}

